##############################################################################################
# ga2geo.pl 
#
#  Script created by Bryce Magee, Teresia Buza, Cathy Gresham
#   
# PURPOSE: The purpose of this script is to transform a gene association file or
#      a GO Retriever file into a GEO formatted file
##############################################################################################
use strict;
use warnings;
use File::stat;
use Time::localtime;
use File::Temp qw/ tempfile /;

##################  Subroutine declarations
sub parseGA($);
sub writeGEO($$);
sub checkOpts;
sub parseObo;

###############  Global variables
my $obofile = "ontology/go-basic.obo";
my $out_GA2GEO_file = "GA2GEO.xls";
my $out_error_file = "GA2GEO_error.xls";
my $ex_obs = 1;
my %goids;
my %obsids;
my %gonames;
my $done;
my $ga;
my $currDate = time();
my $results={};
my $inputfile = "";
my $errorsCnt=0;

############ Get and check user passed in single character switches ##########
# process command line arguments
our $opt_h;
our $opt_e;
our $opt_g;
our $opt_o;
our $opt_i;
our $opt_b;

use Getopt::Std;
getopts('hbe:g:o:i:');

my $printhelp = defined($opt_h);

# check the passed in options
&checkOpts;

#  parse the obo file for list of GO IDs, GO Names
&parseObo;

$results->{geo_info}= &parseGA($inputfile);
if (!$results) {
  die "Parser error. Could not process input file\n. Exiting now.\n\n";
} 
&writeGEO($results, $ex_obs);

print STDERR "\nScript completed. \n\n";
print STDERR "GEO output filename = $out_GA2GEO_file\n\n";

if ($errorsCnt != 0) {
  print STDERR "Error filename = $out_error_file\n\n";
}

################################################################
sub parseGA($){
###############################################################
    my $ga_file = shift;
    local $/ = "\n";

   my %seen;
   open(IN, "<$ga_file");
   while (my $line = <IN>){
      chomp($line);

      my $result = ();
     $result->{error} = [];
      my ($pid, $goid, $evd, $goname, $db_obj_symbol, $db_obj_name, $aspect);
      my $line2 = '';	
	  
	# ignore comment lines
	my $continue=1;
        if ($line =~ /^!.*/) { $continue = 0; }
        if ($line =~ /^Database/) { $continue = 0; } 
        if ($line =~ /^Input_Accession/) { $continue = 0; } 
        if ($line =~ /^Input Accession/) { $continue = 0; } 
   
    my @tmp;
    my $size=1;	
    if ($continue) {	
      @tmp = split("\t", $line,-1);
      $size = $#tmp+1;
      if ($size < 15) {
          push(@{$result->{error}}, { type=> 'Not enough columns',line=>$line });
          $continue = 0;
      }
   } 
    #
    # try to accommodate both a 17 column ga file and the output from GoRetreiver 
    #   which adds an extra column GOTermName in column 6  and input id in col 1
    # 
      if ($continue) {
         if ($size == 17) {
 	    $pid = $tmp[1];
 	    $db_obj_symbol = $tmp[2];
            $goid = $tmp[4];
 	    $evd = $tmp[6];  
            $goname = '';
            $aspect = $tmp[8];
 	    $db_obj_name = $tmp[9];
	}
        else {
 	    $pid = $tmp[2];
 	    $db_obj_symbol = $tmp[3];
            $goid = $tmp[5];
            $goname = $tmp[6];
	    $evd = $tmp[8];      
            $aspect = $tmp[10];
 	    $db_obj_name = $tmp[11];
        }
        $goid =~ s/^\s+//;
        $goid=~ s/\s+$//;
        $evd =~ s/^\s+//;
        $evd =~ s/\s+$//;		
        $aspect =~ s/\r|\n|\t|\f//g;
        $goname =~ s/\r|\n|\t|\f//g;
        $goid =~ s/\r|\n|\t|\f//g;
        $db_obj_name =~ s/\r|\n|\t|\f//g;
        $db_obj_symbol =~ s/\r|\n|\t|\f//g;
        $pid =~ s/\r|\n|\t|\f//g;
        $evd =~ s/\r|\n|\t|\f//g;
	}  # continue -- line had at least 15 columns

    $result->{geo} = [];
    if ($continue) {
        $line2 = "$pid$db_obj_symbol$db_obj_name$goid$goname$aspect$evd";
  	    if (grep { ! $seen{$_} ++ } $line2) {
            push(@{$result->{geo}}, { pid => $pid,
                obj_symbol => $db_obj_symbol,
                obj_name => $db_obj_name,
                goid => $goid,
                goname => $goname,
                aspect => $aspect,
                evd => $evd});
        } #unique lines only
    }
    push(@$ga,$result);
	
}  # while input lines
  
return $ga;
close IN;
}

################################################################
sub writeGEO($$){
###############################################################

my $results = shift;
my $ex_obs= shift;

open(GA2GEO,"> $out_GA2GEO_file") or die "Can not open output text file\n";
open(GA2GEO_ERRORS,"> $out_error_file") or die "Can not open output error file\n";
#
# build a hash for each id, aspect for the ids, names and evds
#
my %id_name = ();
my %pidGO = ();
my %seen = ();

print GA2GEO join ("\t", ('DB_Object_ID','DB_Object_Symbol','DB_Object_Name','GO_ID_CC','GO_NAME_CC','GO_EVD_CC','GO_ID_MF', 'GO_NAME_MF', 'GO_EVD_MF', 'GO_ID_BP','GO_NAME_BP','GO_EVD_BP')), "\n";
# print GA2GEO_ERRORS join ("\t", ('DB_Object_ID','DB_Object_Symbol','DB_Object_Name','GO_ID','GO_NAME','Aspect','EVD','Error')), "\n";

foreach my $result (@{$results->{geo_info}}){
  foreach my $errors (@{$result->{error}}){
       chomp($errors->{line});
       $errors->{line} =~ s/\r\n//g;
       $errors->{line} =~ s/\n//g;
       print GA2GEO_ERRORS "$errors->{line} Error:$errors->{type}\n";
	   
       $errorsCnt++;
  }
  foreach my $geo(@{$result->{geo}}){
    my $pid = $geo->{pid};    
    my $symbol = $geo->{obj_symbol};    
    my $name = $geo->{obj_name};    
    my $go_id = $geo->{goid};
    my $go_name = $geo->{goname};
    my $aspect = uc($geo->{aspect});
    my $evd_code = $geo->{evd};
    #
    # create this variable in case of errors
    # 
    my $line2= join ("\t", ($pid,$symbol,$name,$go_id,$go_name,$aspect,$evd_code));

    my $continue = 1; 

	if ((lc($aspect) ne 'p') && (lc($aspect) ne 'c') && (lc($aspect) ne 'f') ) {
        print GA2GEO_ERRORS "$line2\tError: Invalid Aspect\n";
        $errorsCnt++;
        $continue = 0;
    }
    if ($go_id !~ m/^GO:\d{7}/) {
        print GA2GEO_ERRORS "$line2\tError: Invalid GO ID\n";
        $errorsCnt++;
        $continue = 0;
    }

	if ((defined $ex_obs) && ($ex_obs) && ($obsids{$go_id})) {
        print GA2GEO_ERRORS "$line2\tError: GO ID is flagged as obsolete in the gene ontology file\n";
        $errorsCnt++;
        $continue=0;
    } 
    if (((!defined $go_name) || ($go_name eq ' ' ) || ($go_name eq '' )) ) {
	   if ($gonames{$go_id}) {
	     $go_name=$gonames{$go_id};
	   }
	   else {
          print GA2GEO_ERRORS "$line2\tWarning:GO ID not found in the gene ontology file.\n";
          $errorsCnt++;
          $go_name = "$go_id not found in the gene ontology file";
          $continue=1;	   
	   }
	}	
       
    if ($continue) {
        if (!exists $id_name{$pid}) {
           $id_name{$pid}="$symbol\t$name";
        }
        my $chkLine = "$pid$go_id$go_name$evd_code$aspect";
        if (grep { ! $seen{$_} ++ } $chkLine) {
           if (exists $pidGO{$pid}->{$aspect}) {
              $pidGO{$pid}->{$aspect}->{'id'} .= "; $go_id";
              $pidGO{$pid}->{$aspect}->{'name'} .= "; $go_name";
              $pidGO{$pid}->{$aspect}->{'evd'} .= "; $evd_code";
           }
          else {
              $pidGO{$pid}->{$aspect}->{'id'} =  $go_id;
              $pidGO{$pid}->{$aspect}->{'name'} = $go_name;
              $pidGO{$pid}->{$aspect}->{'evd'} = $evd_code;
          }
        } #not seen line before
    } # continue with creating output hashes -- that is no go id error
  } #foreach geo record
} #foreach result output record        

my ($id, $id_fld, @flds, $id_symbol, $id_name);

my @aspects = ('C','F','P');

foreach $id (sort keys %id_name) {
   $id_fld = $id_name{$id};
   @flds = split(/\t/,$id_fld);
   $id_symbol = $flds[0];
   $id_name= $flds[1];
  # get rid of uninitialized value errors by setting to ''
  if (!defined $id) { $id = ''; }
  if (!defined $id_symbol) { $id_symbol = ''; }
  if (!defined $id_name) { $id_name = ''; }  
  print GA2GEO join ("\t", ($id, $id_symbol, $id_name)),"\t";
   foreach my $asp (@aspects) {
      if (exists $pidGO{$id}->{$asp}) {
	    # get rid of uninitialized value errors by setting to ''
		if (!defined $pidGO{$id}->{$asp}->{'id'}) { $pidGO{$id}->{$asp}->{'id'} = ''; }
		if (!defined $pidGO{$id}->{$asp}->{'name'}) {$pidGO{$id}->{$asp}->{'name'} = ''; }
		if (!defined $pidGO{$id}->{$asp}->{'evd'}) { $pidGO{$id}->{$asp}->{'evd'} = ''; }  
         print GA2GEO "$pidGO{$id}->{$asp}->{'id'}\t";
         print GA2GEO "$pidGO{$id}->{$asp}->{'name'}\t";
         print GA2GEO "$pidGO{$id}->{$asp}->{'evd'}\t";
      }
      else { print GA2GEO "\t\t\t"; }
  }
   print GA2GEO "\n";
}         

 close GA2GEO;
 close GA2GEO_ERRORS;
} # end of writeGEO subroutine

###############################################################################
sub parseObo
###############################################################################
{
 open (OBSGA, $obofile) || die "Cannot open file $obofile: $!\n";
   my ($id, $alt_id, $term, $defCnt, $obsolete, $i);
   $id = $alt_id = $term = '';
   $obsolete = $defCnt = $i = 0;
   while(my $line = <OBSGA>){
     chomp($line);
     if($line =~ m/^\[Term\]/){
        $defCnt++;
        if ($i != 0) {
           if ($id ne '' && $term ne '') {
              $goids{$id}++;
	      $gonames{$id} = $term;			
              if ((defined $alt_id)  && ($alt_id ne '')) {
                     my @alts = split(/\|/,$alt_id);
                     foreach my $alt (@alts) {
                        $goids{$alt}++;
	                $gonames{$alt} = $term;			
                     }
              } # defined alternate ides
              if ((defined $obsolete)  && ($obsolete == 1)) {
                 $obsids{$id}++;
                 if (defined $alt_id && $alt_id ne '') {
                       my @alts = split(/\|/,$alt_id);
                       foreach my $alt (@alts) {
                          $obsids{$alt}++;
                       }
                 } #ALT 
              } # defined obsolete
           } # have an id and term
        } # id not first record
           $i++;
           $id = $alt_id =  $term =  '';
           $obsolete= 0;
     } # Term
     if (($line =~ m/^\[Typedef\]/) ||
           ($line =~ m/^\[Instance\]/)){
           if ($i != 0) {
               if ($id ne '' && $term ne '') {
                 $goids{$id}++;
	         $gonames{$id} = $term;			
                 if ((defined $alt_id)  && ($alt_id ne '')) {
                     my @alts = split(/\|/,$alt_id);
                     foreach my $alt (@alts) {
                        $goids{$alt}++;
	                $gonames{$alt} = $term;			
                     }
                 } # defined alternate ides
                 if ((defined $obsolete)  && ($obsolete == 1)) {
                    $obsids{$id}++;
                    if ((defined $alt_id)  && ($alt_id ne '')) {
                       my @alts = split(/\|/,$alt_id);
                        foreach my $alt (@alts) {
                          $obsids{$alt}++;
                        }
                   }
                 } # defined obsolete
               } # have an id and term
           } # not the first record
           $id = $alt_id =  $term =  '';
           $obsolete= 0;
        }
        if ($line =~ m/^id:\s+(\S+)/)   { $id = $1; }
        if($line =~ m/^def:\s+(.*)/){
                $term= $1;
                $term=~ s/"//g;
	        $term =~ s/^\s+//;
	        $term =~ s/\s+$//;
	        $term =~ s/\r|\n|\t|\f//g;
        }
        if($line =~ m/^alt_id:\s+(.*)/){
          if ($alt_id) { $alt_id .= '|'.$1; }
          else { $alt_id = $1; }
        }
        if($line =~ m/^is_obsolete:\s+(.*)/){ $obsolete = 1; }
   } # while input data
close (OBSGA);

# set the last GO ID if not already processed
if (($id ne '') && ($term ne '')) {
    $goids{$id}++;
    $gonames{$id} = $term;			
   if ((defined $alt_id)  && ($alt_id ne '')) {
       my @alts = split(/\|/,$alt_id);
       foreach my $alt (@alts) {
          $goids{$alt}++;
	  $gonames{$alt} = $term;			
       }
   } # defined alternate ides
   if ((defined $obsolete)  && ($obsolete == 1)) {
       $obsids{$id}++;
       if ((defined $alt_id)  && ($alt_id ne '')) {
           my @alts = split(/\|/,$alt_id);
           foreach my $alt (@alts) {
              $obsids{$alt}++;
           }
        }
    } # defined obsolete
} # have an id and term

} #end of subroutine

###############################################################################
sub checkOpts
###############################################################################
{  
if ($printhelp)
{
print STDERR <<END;
Usage:  perl $0 [-h] [-b] [-g obo_file] [-o GEO file] [-e Error file] -i input file 

Required parameters:
	-i name to a gene association or GORetriever tab-delimited text file.
			  
Optional parameters:
    -h displays this message
    -b do not exclude obsolete go ids. 
		Default is to exclude obsolete go ids.
    -g alternative name to the go-basic.obo file. 
		Default location is ontology/go-basic.obo
    -o alternative name for the GEO output file. 
		Default name is GA2GEO.xls
    -e alternative name for the GEO error file. 
		Default name is GA2GEO_error.xls
             
Examples:
    Convert a file to GEO format 
        % perl GA2GEO.pl -i gene_association.txt

    Convert a file to GEO format designating do not exclude obsolete go ids
        % perl GA2GEO.pl -i gene_association.txt -b 
				  
    Convert a file to GEO format changing location of gene_ontology file
        % perl GA2GEO.pl -i gene_association.txt -g c:\\go-basic.obo
END
    exit;
}

if (!$opt_i) {
  die "Input filename must be provided.\nExiting now.\n\n";
}
else {
  if (!-f $opt_i) {
    die "Input filename $opt_i must exist.\nExiting now.\n\n";
 }
  if (-z $opt_i) {
    die "Input filename $opt_i contains no records.\nExiting now.\n\n";
	}
  if (!-r $opt_i) {
     die "Input filename $opt_i cannot be read.\nExiting now.\n\n";
	 }
} 
if ($opt_g)     {
        $obofile = $opt_g;
}

if (!-f $obofile) {
    die "Ontology filename must exist.\nExiting now.\n\n";
 }
elsif (-z $obofile) {
    die "Ontology filename contains no records.\nExiting now.\n\n";
}
elsif (!-r $obofile) {
     die "Ontology filename cannot be read.\nExiting now.\n\n";
}
my $st = stat($obofile) or die "Cannot stat $obofile: $!";
my $mtime = $st->mtime;
my $age = ($currDate - $mtime) / 86400; # convert seconds to days
if ($age > 30) 
{
  print STDERR "Ontology file is more than 30 days old.  Might want to consider ftping new file\n";  
}

    if ($opt_o)     {
        $out_GA2GEO_file = $opt_o;
    }

    if ($opt_e)     {
        $out_error_file = $opt_e;
    }
	
    if ($opt_i)
    {
        $inputfile = $opt_i;
    }
	
	if ($opt_b) {
	   $ex_obs = 0;
	 }
	else { 
      $ex_obs = 1;
	}   

 #   print STDERR "Input filename = $inputfile\n";
 #   print STDERR "GEO output filename = $out_GA2GEO_file\n";
 #   print STDERR "Error filename = $out_error_file\n";
 #  print STDERR "gene_ontology_edit.obo filename = $obofile\n";
}

